@@ -58,8 +58,8 @@ export BASEDIR=`pwd`
 # Name of the distribution for the image filename (Defaults to 'debian')
 # export CDNAME=debian
 
-# Building squeeze cd set ...
-export CODENAME=squeeze
+# Building wheezy cd set ...
+export CODENAME=wheezy
 
 # By default use Debian installer packages from $CODENAME
 if [ -z "$DI_CODENAME" ]; then
@@ -204,10 +204,18 @@ ATTEMPT_FALLBACK=yes
 # DLDVD:             standard 8.5 GB dual-layer DVD
 # BD:                standard 25 GB blu-ray
 # DLBD:              standard 50 GB dual-layer blu-ray
+# STICK1GB:          1GB USB stick or similar
+# STICK2GB:          2GB USB stick or similar
+# STICK4GB:          4GB USB stick or similar
+# STICK8GB:          8GB USB stick or similar
 # CUSTOM:            up to you - specify a size to go with it (in 2K blocks)
 export DISKTYPE=CD
 #export DISKTYPE=CUSTOM
 #export CUSTOMSIZE=XXXX
+# If you want to over-ride this choice (e.g. to make a larger version of a given disk),
+# you can do the following:
+# export FORCE_CD_SIZE=<type> to change all the sizes in a given run
+# export FORCE_CD_SIZE1=<type> to change the size of disk 1 (only)
 
 # List of languages for which language tasks from tasksel should be
 # included. See tasks/README.tasksel for further info.
@@ -316,6 +324,11 @@ for INCL in $JIGDO_INCLUDE; do
 	JIGDO_OPTS="$JIGDO_OPTS -jigdo-force-md5 $INCL"
 done
 
+# Base link for snapshot.debian.org or similar
+# "SNAPDATETIME" will be replaced at runtime with the correct data
+# Leave this unset to not add this entry
+export SNAPURL=Debian=http://snapshot.debian.org/archive/debian/SNAPDATETIME/
+
 # File with list of packages to include when fetching modules for the
 # first stage installer (debian-installer). One package per line.
 # Lines starting with '#' are comments.  The package order is
